#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : diags.py
# Contents: create diagrams for surrogate data generation chapter
# Authors : Christian Borgelt
# History : 2009.09.02 file created
#           2009.09.03 frame, legend positions improved, shape -> CV
#           2009.09.05 adapted to generalized parameter grids
#           2014.05.27 modules placed in sibling directories
#-----------------------------------------------------------------------
from sys        import argv, path
from os.path    import join
from math       import sqrt
from random     import randint
UTILDIR = join('..', 'util')
if UTILDIR not in path: path.insert(0, UTILDIR)
from id2par     import get_coords, get_args
from surrogates import get_fname
from matplotlib.pyplot import rcParams, figure, savefig

#-----------------------------------------------------------------------

def make_diag (params, dir='results/', summary=True, format='pdf'):
    '''Create diagrams from surrogate data generation results.
    params   figure parameters
    dir      directory with the result files or the summary file
    summary  whether to work with the summary file
    format   format of the figure file (e.g. 'pdf')'''
    d = params[0]; g = params[1:]
    for x in xrange(0,  len(g)):# find the x-coordinate
        if isinstance(g[x][0], ''.__class__):
            if len(g[x][1]) > 1: xr = g[x][1]; break
        elif len(g[x]) > 1:
            xr = range(1,len(g[x])+1); break
    for z in xrange(x+1,len(g)):# find the curve coordinate
        if isinstance(g[z][0], ''.__class__):
            if len(g[z][1]) > 1: zr = g[z][1]; break
        elif len(g[z]) > 1:
            zr = range(1,len(g[z])+1); break
    out = d['fname'] +'.' +format
    print 'file:', out          # get the file name
    if summary:                 # read summary file and
        res = dict()            # store contents in a dictionary
        txt = open(dir+'summary.txt', 'r')
        for line in txt:        # split each line and store
            flds = line.split() # positives and negatives as a pair
            res[flds[0]] = (int(flds[1]), int(flds[2]))
        txt.close()             # close the summary file
    data = dict([(v,[]) for v in zr])
    for i in xrange(get_args([g])): # traverse the param. combs.
        crds  = get_coords([g],i)   # get the grid point coordinates
        args  = get_args([g],crds)  # and the corresponding arguments
        fname = get_fname(**args)   # construct the file name
        if summary:             # get result from summary
            pnc = res[fname]    # retrieve pos./neg. pair
        else:                   # read result from file
            txt = open(dir+fname, 'r')
            pnc = [int(s) for s in txt.read().split()[0:2]]
            txt.close()         # read and evaluate first line
        rate = 100.0*pnc[1 if args['inject'] > 0 else 0] \
             / (pnc[0]+pnc[1])  # compute false pos./neg. rate
        data[crds[z+1]].append(rate)
    rcParams['backend'] = format
    if 'colors' in d:
        c = ['b','r','g','c','m','y']
        m = ['o','o','o','o','o','o']
    else:
        c = ['k','k','k','k','k','k']
        m = ['o','s','^','v','d','.']
    if 'bbox' in d:             # if a bounding box is given,
        bbox = d['bbox']        # set it for the figure to create
        rcParams['figure.subplot.left']   = bbox[0]
        rcParams['figure.subplot.right']  = bbox[1]
        rcParams['figure.subplot.bottom'] = bbox[2]
        rcParams['figure.subplot.top']    = bbox[3]
    if g[x][0] == 'shape':      # transform the shape range
        xr = [1.0/sqrt(v) for v in xr]
    f = figure(); p = f.add_subplot(111)
    for i in xrange(len(zr)):   # traverse and plot the curves
        p.plot(xr, data[zr[i]], color=c[i], marker=m[i], ms=10)
    if 'frame' in d:            # if frame parameters are given
        frame = d['frame']      # adapt the figure frame
        min,max = p.get_xlim(); w = max-min
        min -= frame[0]*w; max += frame[1]*w
        p.set_xlim(min,max)
        min,max = p.get_ylim(); w = max-min
        min -= frame[2]*w; max += frame[3]*w
        p.set_ylim(min,max)
    if 'title'  in d: p.set_title (d['title'],  fontsize='x-large')
    if 'xlabel' in d: p.set_xlabel(d['xlabel'], fontsize='large')
    if 'ylabel' in d: p.set_ylabel(d['ylabel'], fontsize='large')
    if 'legend' in d:           # set the figure title, the labels,
        leg = d['legend']       # and the legend
        loc = dict([('loc',leg[1])] if len(leg) > 1 else [])
        p.legend(leg[0], **loc) # use location if it is given
    savefig(out)                # save the created figure

    #fname = d['fname'] +'.xtb'
    #out = file(fname, 'wt')
    #for i in xrange(len(xr)):
    #    out.write('> ' +str(xr[i]) +' y\n')
    #    for z in zr:
    #        out.write(str(z) +' ' +str(data[z][i]) +'\n')
    #out.close()

    return [zr] +[xr] +[data[zr[i]] for i in xrange(len(zr))]

#-----------------------------------------------------------------------

legend = (r'Trial shuffle',
          r'Spike rand.',
          r'Train dither',
          r'Spike dither',
          r'JISI spike dither',
          r'Spike exchange')
runs   = [1000]
params = [ # figure: FP for set 1 to set 5
           [ { 'fname' : 'fpos',
               'frame' : (0.030,0.030,0.030,0.030),
               #'bbox'  : (0.070,0.998,0.068,0.950),
               'title' : 'false positive rates',
               'xlabel': 'data set',
               'ylabel': 'false positive rate',
               'legend': (legend, (0.637,0.4)), },
             ( { 'rate': 37.75, 'delta': 0,     'shape': 1 },   # set 1
               { 'rate': 0,     'delta': 0,     'shape': 1 },   # set 2
               { 'rate': 37.75, 'delta': 0.235, 'shape': 1 },   # set 3
               { 'rate': 0,     'delta': 0.5,   'shape': 1 },   # set 4
               { 'rate': 0,     'delta': 0.5,   'shape': 3 } ), # set 5
             ('inject', [  0]), # no injected coincidences
             ('method', [0,1,2,3,4,5]),
             ('runs',   runs) ],
           # figure: FN for set 1 to set 5
           [ { 'fname' : 'fneg',
               'frame' : (0.030,0.030,0.030,0.120),
               #'bbox'  : (0.060,0.998,0.068,0.950),
               'title' : 'false negative rates',
               'xlabel': 'data set',
               'ylabel': 'false negative rate',
               'legend': (legend, (0.01,0.655)), },
             ( { 'rate': 37.75, 'delta': 0,     'shape': 1 },   # set 1
               { 'rate': 0,     'delta': 0,     'shape': 1 },   # set 2
               { 'rate': 37.75, 'delta': 0.235, 'shape': 1 },   # set 3
               { 'rate': 0,     'delta': 0.5,   'shape': 1 },   # set 4
               { 'rate': 0,     'delta': 0.5,   'shape': 3 } ), # set 5
             ('inject', [  2]), # injected coincidences with 2Hz
             ('method', [0,1,2,3,4,5]),
             ('runs',   runs) ],
           # figure: FN over injection rate
           [ { 'fname' : 'inject',
               'frame' : (0.030,0.030,0.030,-0.05),
               #'bbox'  : (0.058,0.998,0.068,0.950),
               'title' : 'Effect of injection rate',
               'xlabel': 'Coincidence injection rate',
               'ylabel': 'FN rate (\%)',
               'legend': (legend,), },
             ('rate',   [  0]), # non-stationary within trials
             ('delta',  [0.5]), # non-stationary across trials
             ('shape',  [  3]), # Gamma process with shape 3
             ('inject', [1,1.5,2,2.5,3]),
             ('method', [0,1,2,3,4,5]),
             ('runs',   runs) ],
           # figure: FP over coefficient of variation
           [ { 'fname' : 'shape',
               'frame' : (-0.05,0.020,0.030,0.030),
               #'bbox'  : (0.068,0.998,0.068,0.956),
               'title' : 'Effect of coefficient of variation',
               'xlabel': 'Coefficient of variation',
               'ylabel': 'FP rate (\%)',
               'legend': (legend, (0.637,0.33)) },
             ('rate',   [  0]), # non-stationary within trials
             ('delta',  [0.5]), # non-stationary across trials
             ('shape',  [0.4,0.6,0.8,1,2,3,4,5,8,10]),
             ('inject', [  0]), # no injected coincidences
             ('method', [0,1,2,3,4,5]),
             ('runs',   runs) ],
           # figure: FP over delta (non-stationarity across trials)
           [ { 'fname' : 'delta_fp',
               'frame' : (0.030,0.030,0.030,0.030),
               #'bbox'  : (0.068,0.998,0.068,0.950),
               'title' : 'Effect of non-stationarity across trials',
               'xlabel': 'Amount of non-stationarity (delta)',
               'ylabel': 'FP rate (\%)',
               'legend': (legend, (0.03,0.5)) },
             ('rate',   [  0]), # non-stationary within trials
             ('delta',  [0,0.2,0.4,0.6,0.8,1]),
             ('shape',  [  3]), # Gamma process with shape 3
             ('inject', [  0]), # no injected coincidences
             ('method', [0,1,2,3,4,5]),
             ('runs',   runs) ],
           # figure: FN over delta (non-stationarity across trials)
           [ { 'fname' : 'delta_fn',
               'frame' : (0.030,0.030,0.030,-0.06),
               #'bbox'  : (0.060,0.998,0.068,0.950),
               'title' : 'Effect of non-stationarity across trials',
               'xlabel': 'Amount of non-stationarity (delta)',
               'ylabel': 'FN rate (\%)',
               'legend': (legend, (0.637,0.45)) },
             ('rate',   [  0]), # non-stationary within trials
             ('delta',  [0,0.2,0.4,0.6,0.8,1]),
             ('shape',  [  3]), # Gamma process with shape 3
             ('inject', [  2]), # injected coincidences with 2Hz
             ('method', [0,1,2,3,4,5]),
             ('runs',   runs) ],
         ]

#-----------------------------------------------------------------------

if __name__ == '__main__':
    rcParams['lines.marker']     = 'o'
    rcParams['text.usetex']      = True
    rcParams['font.family']      = 'serif'
    rcParams['legend.labelsep']  = 0
    rcParams['legend.numpoints'] = 1
    rcParams['legend.shadow']    = True
    rcParams['ps.papersize']     = 'auto'

    opts = [ ['dir',     'd', 'results/'], # directory with results
             ['summary', 's', True ],      # wether to use a summary
             ['format',  'f', 'pdf'] ]     # output format
    id = None
    for i in xrange(1,len(argv)):
        if argv[i][0] != '-':
            if id: print 'too many fixed arguments'; exit()
            else:  id = argv[i]; continue
        a = [a for a in opts if a[1] == argv[i][1]]
        if a: a = a[0]          # get the corresponding option tuple
        else: print 'unknown option:', argv[i][:2]; exit()
        v = argv[i][2:]         # get the option argument
        if   isinstance(a[2], (0).__class__): a[2] = int(v)
        elif isinstance(a[2], 0.0.__class__): a[2] = float(v)
        elif isinstance(a[2],True.__class__): a[2] = bool(v)
        else:                                 a[2] = v

    if not id or int(id) <= 0:
        print len(params); exit()
    args = dict([(o[0],o[2]) for o in opts])
    print make_diag(params[int(id)-1], **args)
    
