#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : surrogates.py
# Contents: spike train data and surrogate generation functions
# Authors : Sebastien Louis, Christian Borgelt
# History : 2009.08.?? file created
#           2009.08.17 rate profile creation redesigned
#           2009.08.18 surrogate generation and test added
#           2009.08.19 Gamma process with arbitrary shape added
#           2009.08.20 dither improved, some renaming, function run()
#           2009.08.21 local spike exchange surrogate method ported
#           2009.08.24 spike shifting improved (general ranges)
#           2009.08.25 interspike interval dithering added
#           2009.08.26 permutation generation and test loop limited
#           2009.08.27 safety check for empty rate integral added
#           2009.08.28 surg_exchange improved (better spike assignment)
#           2009.09.02 changes for compatibility with Python 2.5.2
#           2009.09.03 result file explanation added at the end
#-----------------------------------------------------------------------
from sys     import argv, path  # for command line arguments
from os.path import join        # for adding module paths
from string  import ljust       # for string output alignment
from time    import clock, strftime
from random  import seed, shuffle, sample
from math    import ceil, sqrt, exp
from inspect import getargspec

from numpy.random      import seed as npseed
from scipy.signal      import convolve2d as convolve
from matplotlib.pyplot import figure, savefig, show

UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from lists   import integrate, coins, xcoins
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from trains  import randomized, shifted, dithered, isi_dithered
from genpst  import create_rateints, generate_trials

#-----------------------------------------------------------------------
# Auxiliary Function for JISI dithering
#-----------------------------------------------------------------------

def gauss_kernel (size, radius=-1):
    '''Get a Gaussian kernel matrix.
    size:   matrix size (number of rows/columns: 2*size+1)
    radius: radius/standard deviation of the Gaussian function'''
    c = int(size); size = c+c+1; rng = xrange(size);
    s = -1.0/(radius**2) if radius > 0 else -9.0/size**2
    m = [[exp(s*((i-c)**2+(j-c)**2)) for i in rng] for j in rng]
    s = float(sum([sum(row) for row in m]))
    return [[e/s for e in row] for row in m]

#-----------------------------------------------------------------------

def isi_cdfs (trials, size=128, smooth=0, base=0, rng=xrange(1000)):
    '''Compute dither cumulative distribution functions
    via a (smoothed) interspike interval matrix.
    trials: list of spike trains to process
    size:   size of the interspike interval matrix
    smooth: size of the smoothing Gaussian kernel
    base:   value with which to initialize
            the interspike interval matrix
    rng:    range of allowed time bin indices'''
    mat = [[base for k in xrange(size)] for i in xrange(size)]
    for t in trials:            # count spike interval pairs
        t = [rng[0]-1] +t +[rng[-1]+1]
        for i in xrange(1,len(t)-1):
            a = t[i  ]-t[i-1]-1 # distance to previous spike
            b = t[i+1]-t[i  ]-1 # distance to next     spike
            if a+b < size: mat[a][b] += 1
    if smooth > 0:              # smooth the counter matrix
        mat = convolve(mat, gauss_kernel(smooth), mode='same')
    mat = [[sqrt(e) for e in row] for row in mat]
    mat = [[0] +integrate([mat[i-k][k] for k in xrange(i+1)])
           for i in xrange(size)]
    return mat                  # collect intervall pairs with same sum

#-----------------------------------------------------------------------
# Surrogate Creation Functions
#-----------------------------------------------------------------------

def surg_shuffle (trials):
    '''Create a shuffled copy of the trials (spike trains) of a neuron.
    trials: trials (list of spike trains) to shuffle'''
    out = [t for t in trials]; shuffle(out)
    return out                  # shuffle a copy of the list of trials

#-----------------------------------------------------------------------

def surg_randomize (trials, rng=xrange(1000)):
    '''Create a randomized copy of the trials (spike trains)
    of a neuron (spikes relocated to random positions).
    trials: trials (list of spike trains) to randomize
    rng:    range of allowed time bin indices'''
    return [sample(rng, len(t)) for t in trials]
    # return [randomized(t,rng) for t in trials]

#-----------------------------------------------------------------------

def surg_shift (trials, shift=20, rng=xrange(1000)):
    '''Create a randomly shifted copy of the trials
    (spike trains) of a neuron.
    trials: trials (list of spike trains) to shift
    dither: maximum amount by which to shift the spikes
    rng:    range of allowed time bin indices'''
    return [shifted(t,shift,rng) for t in trials]

#-----------------------------------------------------------------------

def surg_dither (trials, dither=20, rng=xrange(1000)):
    '''Create a dithered copy of the trials (spike trains) of a neuron.
    trials: trials (list of spike trains) to dither
    dither: maximum amount by which to shift a spike
    rng:    range of allowed time bin indices'''
    return [dithered(t,dither,rng) for t in trials]

#-----------------------------------------------------------------------

def surg_jisidither (trials, cdfs, isid=20, rng=xrange(1000)):
    '''Create a copy of the trials with dithered interspike intervals.
    trials: trials (list of spike trains) to shuffle
    isid:   maximum amount by which to shift a spike
    size:   size of the interspike interval cdf table
    rng:    range of allowed time bin indices'''
    return [isi_dithered(t,cdfs,isid,rng) for t in trials]

#-----------------------------------------------------------------------

def surg_exchange (trials, window=20, rng=xrange(1000)):
    '''Create a copy of the trials with spikes
    randomly reassigned between trials in time windows.
    trials: trials (list of spike trains) to process
    window: size of the windows
    rng:    range of allowed time bin indices'''
    window = int(window)        # ensure an integer window size
    xrt = xrange(len(trials))   # get often needed ranges
    xrw = xrange(int(ceil((rng[-1]+1.0)/window)))
    mat = [[[] for t in xrt] for w in xrw]
    out = [ [] for t in xrt]    # initialize windows and output
    for t in xrt:               # split all trials into windows:
        for s in trials[t]:     # collect spikes per window per trial
            mat[s/window][t].append(s)
    for w in xrw:               # traverse the windows
        wls = mat[w]            # get the windows for all trials
        tms = [[t,len(wls[t]),[]] for t in xrt if len(wls[t]) > 0]
        spikes = []             # collect all spikes in the windows
        for t in tms: spikes.extend(wls[t[0]])
        n = len(spikes)         # get the total number of spikes
        if n <= 0: continue     # if there is none, skip the window
        i = 1; r = min(n,4)     # init. spike index and perm. counter
        while i < n and r > 0:  # permutation generate and test loop
            shuffle(spikes)     # generate next spike permutation
            i = 0               # test the generated permutation
            for t in tms:       # traverse the new spike assignments
                if len(set(spikes[i:i+t[1]])) != t[1]: r -= 1; break
                i += t[1]       # check for spike uniqueness
        if r > 0:               # if a valid permutation was found,
            i = 0               # append the modified windows
            for t in tms:       # to the output trials
                out[t[0]].extend(spikes[i:i+t[1]])
                i += t[1]
            continue            # continue with the next window
        # The following way of generating an appropriate reassignment
        # of the spikes is slower (for a single run), but much less
        # likely to fail than the permutation method above. Therefore
        # trying the permutation method first and falling back on the
        # following method if it fails (4 unsuccessful trials to find
        # a valid permutation) yields the shortest execution times.
        tms.sort(lambda a,b: b[1]-a[1])
        off = w*window; r = 16  # sort with decreasing number of spikes
        while n > 0 and r > 0:  # permutation generation loop
            sms = [[] for i in xrange(window)]
            for s in spikes: sms[s-off].append(s)
            for t in tms:       # compute the spike time marginals
                pool = []       # collect the assignable spikes
                for s in sms: pool.extend(s)
                n = t[1]; t[2] = []
                while n > 0 and len(pool) >= n:
                    x = set(sample(pool,n))
                    t[2].extend([s for s in x])
                    n -= len(x) # select, assign and delete spikes
                    if n > 0: pool = [s for s in pool if s not in x]
                if n > 0: r -= 1; break
                for s in t[2]:  # delete selected/assigned spikes
                    del sms[s-off][-1]
        if r > 0:               # if a valid permutation was found,
            for t in tms:       # append the modified windows
                out[t[0]].extend(t[2])
        else:                   # if no valid assignment was found,
            for t in tms:       # append the original windows
                out[t[0]].extend(wls[t[0]])
    return out                  # return the modified trials

#-----------------------------------------------------------------------

surrogate_methods = [ ('shuffle',    surg_shuffle),
                      ('randomize',  surg_randomize),
                      ('shift',      surg_shift),
                      ('dither',     surg_dither),
                      ('jisidither', surg_jisidither),
                      ('exchange',   surg_exchange), ]

#-----------------------------------------------------------------------

def test_surrogates (rate=40, delta=0, shape=1, trials=40,
                     inject=1, jitter=1,
                     dither=20, shift=20, isid=20, window=20,
                     surgs=1000, salt=None):
    '''Test surrogate generation procedures by generating a trial
    with two neurons and then generating surrogates for it.
    rate:   rate of a stationary process
            or <= 0 for a special rate profile
    delta:  delta for the rate modification between trials
    shape:  shape parameter of the process (Poisson or Gamma)
    trials: number of trials (number of spike trains per neuron)
    inject: coincidence injection rate
    jitter: jitter for coincidence spikes
    dither: dither for spike time dithering
    shift:  shift for spike train shifting
    isid:   dither for interspike interval dithering
    window: window size for spike exchanging
    surgs:  number of surrogates to generate
    salt:   value with which to seed the random number generator'''
    if salt != 0:               # seed random number generators
        seed(salt); npseed(salt)

    print 'generating', trials, 'trials for two neurons:',
    t = clock()                 # create rate profile
    ris = create_rateints(rate, delta)
    mat = generate_trials(2, trials, ris, inject, jitter)
    tol = [0,1]                 # generate trials
    acc = [sum([xcoins(a,b,j,True) for a,b in zip(mat[0],mat[1])])
           for j in tol]        # compute actual coincidence counts
    print '{0:5.2f}s'.format(clock()-t)

    m = len(surrogate_methods)  # loop over all surrogate methods
    cdfs = isi_cdfs(mat[0]) if m > 4 else []
    args = {'dither': dither, 'shift': shift, 'isid': isid,
            'window': window, 'cdfs': cdfs, 'rng': xrange(len(ris[0]))}
    scc = []                    # init. the coins. counter list
    for name, method in surrogate_methods[0:m]:
        print ljust(name,23) +' (' +str(surgs) +' times):',
        t = clock()             # filter function arguments
        spec = getargspec(method)[0][1:]
        curr = dict([(k,args[k]) for k in spec if k in args])
        cc   = [[] for j in tol]
        for i in xrange(surgs): # surrogate creation loop
            s = [sorted(x) for x in method(mat[0], **curr)]
            for j in tol:       # for each tolerance value
                c = sum([xcoins(a,b,j,True) for a,b in zip(s,mat[1])])
                cc[j].append(c) # count coincidences in all trials
        scc.append(cc)          # collect the coincidence counts
        print '{0:5.2f}s '.format(clock()-t),
        for j in tol:           # compute and print p-values
           p = len([x for x in cc[j] if x >= acc[j]]) /(surgs+0.0)
           print '{0:5.3f}'.format(p),
        print ''                # terminate the output line

    fig = figure(1); diags = range(2*m); hists = range(2*m)
    scc = [c[0] for c in scc] + [c[1] for c in scc]
    for i in xrange(2*m):       # create the histograms
        diags[i] = fig.add_subplot(2,m,i+1)
        hists[i] = diags[i].hist(scc[i], max(scc[i])-min(scc[i]))
    xmax = max([h[1][-1] for h in hists] +[a+1 for a in acc])
    ymax = max(max(h[0]) for h in hists)
    acc  = [[(a,a),(0,ymax),'r'] for a in acc]
    ymax = int(ymax *1.05 +0.5)
    diags[0].set_ylabel('jitter = 0')
    diags[m].set_ylabel('jitter = 1')
    for i in xrange(m):         # set diagram names
        diags[i  ].set_title(surrogate_methods[i][0])
        diags[i  ].plot(*acc[0])# plot the coincidence count
        diags[i+m].plot(*acc[1])# in the original spike trains
    for d in diags:             # unify the diagram limits
        d.set_xlim(0,xmax)
        d.set_ylim(0,ymax)
    savefig('test.pdf', format='pdf')
    show()                      # show the created diagrams

#-----------------------------------------------------------------------
# Worker Function for Experiments
#-----------------------------------------------------------------------

def get_fname (rate=40, delta=0, shape=1, trials=40,
               inject=1, jitter=1,
               method=2, shift=20, dither=20, isid=20, window=20,
               surgs=1000, cctol=1, runs=100):
    '''Get the name of the output file for a parameter combination.
    rate:   rate of a stationary process
            or <= 0 for a special rate profile
    delta:  delta for the rate modification between trials
    shape:  shape parameter of the process (Poisson or Gamma)
    trials: number of trials (number of spike trains per neuron)
    inject: coincidence injection rate
    jitter: jitter for coincidence spikes
    method: surrogate generation method
    shift:  shift for spike train shifting
    dither: dither for spike time dithering
    isid:   dither for interspike interval dithering
    window: window size for spike exchanging
    surgs:  number of surrogates to generate
    cctol:  tolerance for coincidence counting
    runs:   number of runs (repetitions)'''
    return 'r' +str(rate)   \
         + 'd' +str(delta)  \
         + 's' +str(shape)  \
         + 't' +str(trials) \
         + 'i' +str(inject) \
         + 'j' +str(jitter) \
         + 'm' +str(method) \
         + 'u' +str(surgs)  \
         + 'h' +str(shift)  \
         + 'e' +str(dither) \
         + 'z' +str(isid)   \
         + 'w' +str(window) \
         + 'c' +str(cctol)  \
         + 'x' +str(runs)   \
         + '.txt'
    # The following requires Python 2.6, does not work in 2.5.2.
    # return 'r' +'{0:g}'.format(rate)   \
    #      + 'd' +'{0:g}'.format(delta)  \
    #      + 's' +'{0:g}'.format(shape)  \
    #      + 't' +'{0:d}'.format(trials) \
    #      + 'i' +'{0:g}'.format(inject) \
    #      + 'j' +'{0:d}'.format(jitter) \
    #      + 'm' +'{0:d}'.format(method) \
    #      + 'u' +'{0:d}'.format(surgs)  \
    #      + 'h' +'{0:d}'.format(shift)  \
    #      + 'e' +'{0:d}'.format(dither) \
    #      + 'z' +'{0:d}'.format(isid)   \
    #      + 'w' +'{0:d}'.format(window) \
    #      + 'c' +'{0:d}'.format(cctol)  \
    #      + 'x' +'{0:d}'.format(runs)   \
    #      + '.txt'

#-----------------------------------------------------------------------

def run_exps (rate=40, delta=0, shape=1, trials=40,
              inject=1, jitter=1,
              method=2, shift=20, dither=20, isid=20, window=20,
              surgs=1000, cctol=1, runs=100, salt=None, verbose=10):
    '''Run a set of surrogate generation experiments.
    rate:   rate of a stationary process
            or <= 0 for a special rate profile
    delta:  delta for the rate modification between trials
    shape:  shape parameter of the process (Poisson or Gamma)
    trials: number of trials (number of spike trains per neuron)
    inject: coincidence injection rate
    jitter: jitter for coincidence spikes
    method: surrogate generation method
    shift:  shift for spike train shifting
    dither: dither for spike time dithering
    isid:   dither for interspike interval dithering
    window: window size for spike exchanging
    surgs:  number of surrogates to generate
    cctol:  tolerance for coincidence counting
    runs:   number of runs (repetitions)
    salt:   value with which to seed the random number generator'''
    if salt != 0:               # seed random number generators
        seed(salt); npseed(salt)

    ris  = create_rateints(rate, delta)
    meth = surrogate_methods[method][1]
    spec = getargspec(meth)[0][1:]
    args = {'dither': dither, 'shift': shift, 'isid': isid,
            'window': window, 'rng': xrange(len(ris[0]))}
    args = dict([(k,args[k]) for k in spec if k in args])

    results = []                # initialize the result
    for r in xrange(runs):      # execute runs
        if r % verbose == 0:    # print a log message
            print(strftime('%Y.%m.%d %H:%M:%S runexp ')+str(r))
        mat = generate_trials(2, trials, ris, shape, inject, jitter)
        acc = sum([xcoins(a,b,cctol,True)
                   for a,b in zip(mat[0],mat[1])])
        scc = []                # count actual coincidences
        if meth == surg_jisidither:
            args['cdfs'] = isi_cdfs(mat[0])
        for i in xrange(surgs): # create surrogates
            s = [sorted(x) for x in meth(mat[0], **args)]
            c = sum([xcoins(a,b,cctol,True) for a,b in zip(s,mat[1])])
            scc.append(c)       # collect coincidence counts
        p = len([x for x in scc if x >= acc]) /(surgs+0.0)
        results.append((p, acc, scc))
    print(strftime('%Y.%m.%d %H:%M:%S runexp ')+str(runs)+' done.')
    r = len([x for x in results if x[0] <= 0.01])
    if inject == 0: print str(100.0*      r /runs) +'% false positives'
    else:           print str(100.0*(runs-r)/runs) +'% false negatives'

    fname = get_fname(rate, delta, shape, trials, inject, jitter,
                      method, dither, shift, isid, window,
                      surgs, cctol, runs)
    out = file(fname, 'wt')
    out.write(str(r) +' ' +str(runs-r) +'\n')
    for r in results:
        out.write(str(r[0]) +' ' +str(r[1]) +' ')
        for c in r[2]: out.write(' ' +str(c))
        out.write('\n')
    out.close()                 # 'with' automatically closes the file

    # The following requires Python 2.6, as 2.5.2 does not know 'with'.
    # with open(fname, 'w') as out:
    #     out.write(str(r) +' ' +str(runs-r) +'\n')
    #     for r in results:
    #         out.write(str(r[0]) +' ' +str(r[1]) +' ')
    #         for c in r[2]: out.write(' ' +str(c))
    #         out.write('\n')

#-----------------------------------------------------------------------
# Main
#-----------------------------------------------------------------------
if __name__ == '__main__':
    opts = [ ['rate',   'r',   40.0],
             ['shape',  's',    1.0],
             ['delta',  'd',    0.0],
             ['trials', 't',   40],
             ['inject', 'i',    1.0],
             ['jitter', 'j',    1],
             ['method', 'm',    0],
             ['shift',  'h',   20],
             ['dither', 'e',   20],
             ['isid',   'z',   20],
             ['window', 'w',   20],
             ['surgs',  'u', 1000],
             ['cctol',  'c',    1],
             ['runs',   'x',   10],
             ['salt',   'S',    0] ]
    fixed = None
    for i in xrange(1,len(argv)):
        if argv[i][0] != '-':
            if fixed: print 'too many fixed arguments'; exit()
            else:     fixed = argv[i]; continue
        a = [a for a in opts if a[1] == argv[i][1]]
        if a: a = a[0]          # get the corresponding option tuple
        else: print 'unknown option:', argv[i][:2]; exit()
        v = argv[i][2:]         # get the option argument
        if a[0] == 'method':    # handle surrogate method specially
            m = [i for i in xrange(len(surrogate_methods))
                 if surrogate_methods[i][0] == v]
            if m: a[2] = m[0]; continue
        if   isinstance(a[2],(0).__class__): a[2] = int(v)
        elif isinstance(a[2],0.0.__class__): a[2] = float(v)
        else:                                a[2] = v

    if fixed:                   # if a fixed argument is given
        spec = getargspec(test_surrogates)[0]
        args = dict([(o[0],o[2]) for o in opts])
        args = dict([(k,args[k]) for k in spec if k in args])
        test_surrogates(**args) # execute a test procedure
    else:                       # if there are only options
        print 'running with these values:'
        for a in opts: print ljust(a[0],6) +' = ' +str(a[2])
        run_exps(**dict([(a,c) for a,b,c in opts]))

#-----------------------------------------------------------------------
# naming of the result files
#-----------------------------------------------------------------------
# example: r0d0s1t40i2j1m0u1000h20e20z20w20c1x100.txt
#
# r: rate (profile) of the process
#     0: non-stationary rate, two burst profile
#    >0: stationary rate, for example r40: 40Hz
# d: delta factor
#     0: stationary across trials
#    >0: non-stationary across trials
# s: shape parameter
#     1: Poisson
#    >1: Gamma
# t: number of trials
#    value: number of trials per neuron
# i: injection rate
#     0: no coincidences ejected
#    >0: coincidences injected with given rate
# j: jitter for injected coincidences
#    value: number of time bins
# m: surrogate generation method
#    0: trial shuffling
#    1: spike time randomization/shuffling
#    2: spike train shifting/rotating
#    3: spike time dithering
#    4: interspike interval shuffling
#    5: spike exchange method
# u: number of surrogates to generate
#    value: number of surrogate data sets
# h: shift distance (train dither)
#    value: maximum number of time bins by which to shift a spike
# e: dither width (spike dither)
#    value: maximum number of time bins by which to shift a spike
# z: dither width for JISI dithering
#    value: maximum number of time bins by which to shift a spike
# w: windows size for spike exchanges
#    value: windows size in number of time bins
# c: tolerance for coincidence counting
#    value: maximum tolerated difference in number of time bins
# x: number of runs
#    value: number of original spike trains to generate
#
#-----------------------------------------------------------------------
# contents of result files
#-----------------------------------------------------------------------
# first line:
#    number of positive and negative cases / runs
#    (1% confidence level)
# each following line (as many as the sum of the values in first line):
#    p-value
#    actual coincidence count
#    list of coincidence counts in surrogates
