#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : genpst.py
# Contents: parallel spike train data generation functions
# Authors : Sebastien Louis, Christian Borgelt
# History : 2009.08.?? file created
#           2009.08.17 rate profile creation redesigned
#           2009.08.19 Gamma process with arbitrary shape added
#           2009.08.27 safety check for empty rate integral added
#           2009.09.02 changes for compatibility with Python 2.5.2
#           2010.06.22 data generation functions made a separate file
#-----------------------------------------------------------------------
from sys     import argv        # for command line arguments
from time    import clock       # for run time measurements
from bisect  import bisect      # for binary search in sorted lists
from random  import random, randint, sample
from math    import exp, log
from inspect import getargspec

from scipy.stats       import poisson, expon, gamma
from matplotlib.pyplot import figure, savefig, show

from lists   import integrate, xcoins
from trains  import dithered

#-----------------------------------------------------------------------
# Rate Profile and Sample Generation Functions
#-----------------------------------------------------------------------

def burst_profile (period, mode, ratio=1.25, bps=1000.0):
    '''Create a burst rate profile.
period: length of the profile in seconds
mode:   position of the mode, in seconds from start
ratio:  ratio of the time constants
bps:    time bins per second
returns a list of firing rates, one per time bin'''
    mode = float(mode)*bps      # compute the mode bin offset
    a    = mode*(ratio-1.0)/(ratio*log(ratio))
    b    = ratio*a              # compute the time constants
    n    = int(period*bps+(1.0-1e-9))       # number of time bins
    s    = 1.0/(exp(-mode/a)-exp(-mode/b))  # scaling factor
    return [s*(exp(-i/a)-exp(-i/b)) for i in xrange(n)]

#-----------------------------------------------------------------------

def create_profile (base0, base1, base2, scale1, scale2, mode,
                    delta=0.0):
    '''Generate a complex rate profile with two bursts.
base0:  base rate before first  burst (0.0 to 0.1)
base1:  base rate for    first  burst (0.1 to 0.6)
base2:  base rate for    second burst (0.6 to 1.0)
scale1: scaling factor for first  burst (relative to mode)
scale2: scaling factor for second burst (relative to mode)
delta:  delta for profile modification
returns a list of firing rates, one per time bin'''
    if delta < -1: delta = -1   # check and adapt delta
    scale1 *= 1.0+delta; s1 = scale1+(base1-base0)
    scale2 *= 1.0+delta; s2 = scale2+(base2-base1)
    m = int(mode*1000)          # get scaling factors and mode offset
    p = burst_profile(0.5,mode) # create burst profile and collate
    return [0.0     +base0 for x in xrange(100)] \
         + [x*s1    +base0 for x in p[0:m]]      \
         + [x*scale1+base1 for x in p[m:500]]    \
         + [x*s2    +base1 for x in p[0:m]]      \
         + [x*scale2+base2 for x in p[m:400]]

#-----------------------------------------------------------------------

def create_rateints (rate=40, delta=0):
    '''Create a list of rate integrals for a given rate and delta.
    rate:  rate of a stationary process
           or <= 0 for a special rate profile
    delta: delta for profile modification
    returns a list of rate integrals'''
    if delta > 1: delta = 1     # check and adapt delta
    if rate > 0:                # if stationary firing rate
        if delta <= 0:          # if only one rate profile
              prof = [ [rate             for i in xrange(1000)] ]
        else: prof = [ [rate*(1.0-delta) for i in xrange(1000)],
                       [rate*(1.0+delta) for i in xrange(1000)] ]
    else:                       # if non-stationary firing rate
        if delta <= 0:          # if only one rate profile
              prof = [ create_profile(10,30,10, 80,50, 0.05) ]
        else: prof = [ create_profile(10,30,10, 80,50, 0.05, -delta),
                       create_profile(10,30,10, 80,50, 0.05, +delta) ]
    return [integrate(p) for p in prof]  # integrate all profiles

#-----------------------------------------------------------------------

def sample_profile (rateint, shape=1, bps=1000.0):
    '''Sample a Poisson or Gamma process with a given rate integral.
    rateint: rate integral of the process
    shape:   shape parameter of the process (1: Poisson, > 1: Gamma)
    bps:     time bins per second
    returns a list of spike times'''
    k = rateint[-1]             # get the range of time bin indices
    if k <= 0: return []        # and check for a vanishing rate
    if shape <= 0: shape = 1    # correct an illegal shape parameter
    s = int(shape)              # and check for a fractional part
    if shape != s:              # if shape parameter is not integer
        i = bps*expon.rvs()     # start with a Poisson interval
        s = bps/shape; o = []   # get scale and init. output train
        while i <= k:           # add Gamma steps while in range
            o.append(bisect(rateint, i))
            i += s*gamma.rvs(shape)
        return o                # warp and return the result
    n = poisson.rvs(shape*k/(bps+0.0))
    t = sorted([random() for i in xrange(n)])
    # t = sorted(sample(xrange(k), int(n)))
    # The sample function ensures distinct elements, but is slower.
    return [bisect(rateint, k*t[i])
            for i in xrange(randint(0,s-1),n,s)]
    # It is important to use a random starting point for the selection,
    # because otherwise the starts of Gamma processes are too similar.

    # Possible alternative method for generating a gamma process
    # with a fractional shape parameter > 1 (to be tested):
    # i = int(shape*random())     # choose a random start index
    # o = []; f = shape-s         # init. the output spike train
    # while i < n:                # select with a variable step width
    #     o.append(bisect(rateint, k*t[i]))
    #     i += s+1 if random() < f else s
    # return o                    # (range exprs. cannot handle this)
    # Note that this version for a fractional shape parameter can also
    # handle integer shape values. However, it is somewhat slower.

#-----------------------------------------------------------------------

def test_sampling (delta=0.5):
    '''Test sampling procedures by drawing 1000 sample 
    from a rate profile with two bursts.
    delta: delta for profile modification'''
    prof = create_profile(10,30,10, 80,50, 0.05)
    low  = create_profile(10,30,10, 80,50, 0.05, -delta)
    high = create_profile(10,30,10, 80,50, 0.05, +delta)
    mean = [0.5*(x+y) for x,y in zip(high,low)]
    print 'diff:', max(abs(x-y) for x,y in zip(prof,mean))
    print 'sampling rate profile 1000 times:',
    t = clock()
    s = []; pint = integrate(prof)
    for i in xrange(1000):      # aggregate 1000 samples
        s.extend(sample_profile(pint,1))
    print '{0:5.2f}s'.format(clock()-t)
    diag = figure(1).add_subplot(111)
    diag.plot(prof); diag.plot(low); diag.plot(high)
    # diag.plot(mean)             # identical to prof
    h = diag.hist(s, len(prof)) # draw samples as a histogram
    diag.set_title('rate profiles and aggregate of 1000 samples')
    savefig('test.pdf', format='pdf')
    show()

#-----------------------------------------------------------------------
# Trial Generation Functions
#-----------------------------------------------------------------------

def generate_trials (nNeurons, nTrials, rateint, shape=1,
                     inject=0, jitter=0, bps=1000.0, dedup=True):
    '''Generate a set of spike trains for a set of neurons.
    nTrials:  number of trials to generate
    nNeurons: number of neurons for which to generate spike trains
    rateint:  either a single rate integral or
              a list or tuple of rate integrals to choose from
    inject:   rate of injected coincidences (<= 0: no coincidences)
    jitter:   jitter to apply to injected coincidences
    bps:      time bins per second
    This function returns a list, each element of which is a list of
    trials for one neuron. A trial for one neuron is a (binned) spike
    train, that is, a list of spike time bin indices. The result may
    also be seen as a matrix with the rows (first index) corresponding
    to neurons, the columns (second index) corresponding to trials.'''
    if  not isinstance(rateint[0], [].__class__) \
    and not isinstance(rateint[0], ().__class__):
        rateint = [rateint]     # rateint must be a list of integrals
    mat = [[] for i in xrange(nNeurons)]
    xrt = xrange(nTrials); n = len(rateint)
    for col in xrt:             # trial creation loop (sampling)
        r = rateint[int(n*random())]
        for row in mat:         # choose rate profile and sample
            row.append(sample_profile(r, shape, bps))
    if inject != 0:             # if to inject coincidences
        n = len(rateint[0])     # get the number of time bins
        r = xrange(n)           # and the corresponding range
        k = abs(inject)*n/(bps+0.0) # compute the (avg.) coin. count
        for col in xrt:         # coincidence injection loop
            # c = poisson.rvs(k) if inject > 0 else \
            # for compatibility with older versions of scipy:
            c = poisson.rvs(k,size=1)[0] if inject > 0 else \
                int(k+1) if random() < k-int(k) else int(k)
            # t = [randint(0,n-1) for i in xrange(c)]
            t = sample(r,c)
            # Since the injection rate can be expected to be low,
            # the slower sample function is acceptable here.
            # However, the other form is compatible with older
            # versions of scipy, even if the first form of calling
            # poisson.rvs is used as xrange(c) works for arrays c.
            for row in mat:     # create and add coincidence spikes
                row[col].extend(t if jitter <= 0
                                else dithered(t,jitter,r))
    if dedup:                   # if to remove duplicates
        for col in xrt:         # remove duplicates and sort spikes
            for row in mat:     # in all trials for all neurons
                row[col] = sorted([s for s in set(row[col])])
    else:                       # if not to remove duplicates
        for col in xrt:         # only sort the spikes
            for row in mat:     # in all trials for all neurons
                row[col].sort()
    return mat                  # return the created trial matrix

#-----------------------------------------------------------------------

def test_trialgen (rate=0, delta=0, inject=1, jitter=1):
    '''Test trial generation by generating 1000 trials for 2 neurons.
    rate:   rate of a stationary process
            or <= 0 for a special rate profile
    delta:  delta for the rate modification between trials
    inject: coincidence injection rate
    jitter: jitter for coincidence spikes'''

    print 'generating 1000 trials for two neurons:     ',
    t   = clock()               # generate 2x1000 spike trains
    ris = create_rateints(rate, delta)
    mat = generate_trials(2, 1000, ris, inject, jitter)
    print '{0:5.2f}s'.format(clock()-t)

    print 'intersecting the spike trains of the trials:',
    t = clock(); cc = [[],[]]   # intersect the spike trains
    for a,b in zip(mat[0],mat[1]):
        cc[0].append(len(set(a).intersection(b)))
        cc[1].append(xcoins(a,b,1,True))
    print '{0:5.2f}s'.format(clock()-t)

    fig = figure(1); diags = [0,1]; hists = [0,1]
    for i in [0,1]:             # generate the histograms
        diags[i] = fig.add_subplot(2,1,i+1)
        hists[i] = diags[i].hist(cc[i], max(cc[i])-min(cc[i]))
        diags[i].set_ylabel('tolerance = ' +str(i))
    xmax = max(h[1][-1]  for h in hists)
    ymax = max(max(h[0]) for h in hists)
    ymax = int(ymax*1.05 +0.5)
    for d in diags:             # unify the diagram limits
        d.set_xlim(0,xmax)
        d.set_ylim(0,ymax)
    diags[0].set_title('number of trials over coincidence counts')
    savefig('test.pdf', format='pdf')
    show()                      # set title and show diagrams

#-----------------------------------------------------------------------

if __name__ == '__main__':
    opts = [ ['rate',   'r',   40.0],
             ['shape',  's',    1.0],
             ['delta',  'd',    0.0],
             ['trials', 't',   40],
             ['inject', 'i',    1.0],
             ['jitter', 'j',    1] ]
    fixed = None
    for i in xrange(1,len(argv)):
        if argv[i][0] != '-':
            if fixed: print 'too many fixed arguments'; exit()
            else:     fixed = argv[i]; continue
        a = [a for a in opts if a[1] == argv[i][1]]
        if a: a = a[0]          # get the corresponding option tuple
        else: print 'unknown option:', argv[i][:2]; exit()
        v = argv[i][2:]         # get the option argument
        if a[0] == 'method':    # handle surrogate method specially
            m = [i for i in xrange(len(surrogate_methods))
                 if surrogate_methods[i][0] == v]
            if m: a[2] = m[0]; continue
        if   isinstance(a[2],(0).__class__): a[2] = int(v)
        elif isinstance(a[2],0.0.__class__): a[2] = float(v)
        else:                                a[2] = v
    if not fixed: fixed = 0
    sfns = [test_sampling, test_trialgen]
    fn   = sfns[int(fixed)]
    spec = getargspec(fn)[0]
    args = dict([(o[0],o[2]) for o in opts])
    args = dict([(k,args[k]) for k in spec if k in args])
    fn(**args)              # execute a test procedure
